@@ -37,7 +37,7 @@ class Image(np.ndarray):
     def __new__(cls, im: Union[str, np.ndarray] = None, color_mode: str = 'BGR', shape=None, dtype=None,
                 **kwargs) -> 'Image':
         if im is None:
-            data = np.zeros(kwargs.get('shape', (480, 640)), dtype=dtype or np.uint8)
+            data = np.zeros(kwargs.get('shape', (480, 640, 3)), dtype=dtype or np.uint8)
         else:
             assert isinstance(color_mode, str), f'channels {color_mode} is not valid.'
             if isinstance(im, str):
@@ -60,7 +60,7 @@ class Image(np.ndarray):
                 io_buf.seek(0)
                 data = np.reshape(np.frombuffer(io_buf.getvalue(), dtype=np.uint8),
                                   newshape=(int(im.bbox.bounds[3]), int(im.bbox.bounds[2]), -1))
-            elif isinstance(im, np.ndarray):
+            elif isinstance(im, (np.ndarray, Image)):
                 data = im.copy()
             else:
                 raise TypeError(f'{type(im)} is not valid.')
@@ -266,8 +266,7 @@ class Image(np.ndarray):
             self._color_mode = 'BGR'
             self[:, ] = cv2.cvtColor(self, cv2.COLOR_RGBA2BGR)
         elif self._color_mode == 'BGRA':
-            self._color_mode = 'BGR'
-            self[:, ] = cv2.cvtColor(self, cv2.COLOR_BGRA2BGR)
+            return Image(cv2.cvtColor(self.copy(), cv2.COLOR_BGRA2BGR))
         return self
 
     def resize(self, shape: Union[tuple, list], keep_scale: bool = False) -> Image:
@@ -418,8 +417,8 @@ class Image(np.ndarray):
         text = str(text)
         text_size, _ = cv2.getTextSize(text, font, font_scale, font_thickness)
         if text_size[0] > self.width:
-            font_scale = min(text_size[0] / (self.width - (self.width * 0.2)), font_scale)
-
+            # font_scale = min(text_size[0] / (self.width - (self.width * 0.2)), font_scale)
+            font_scale = self.width / text_size[0]
         text_size, _ = cv2.getTextSize(text, font, font_scale, font_thickness)
         text_size = text_size[0], int(text_size[1] * 1.5)
         k = int(self.min_len * 0.03)
@@ -464,7 +463,7 @@ class Image(np.ndarray):
 
     def laplacian_pyramid(self, levels=None) -> Sequence[Image]:
         levels = levels or self.pyramid_levels
-        gaussian_pyramid = self.gaussian_pyramid(levels)
+        gaussian_pyramid = self.gaussian_pyramid(levels=levels)
         pyramid = []
         for i in range(levels, 0, -1):
             GE = cv2.pyrUp(gaussian_pyramid[i], dstsize=gaussian_pyramid[i - 1].shape[::-1][1:])
@@ -479,8 +478,8 @@ class Image(np.ndarray):
 
 
 class VideoWriter:
-    def __init__(self, p, width=None, height=None, fps=30):
-        self._fourcc = cv2.VideoWriter_fourcc(*'XVID')
+    def __init__(self, p, fourcc=None, width=None, height=None, fps=30):
+        self._fourcc = -1 if fourcc is None else cv2.VideoWriter_fourcc(*fourcc) if isinstance(fourcc, str) else fourcc
         self._height, self._width = width, height
         self._path = p
         self.__writer = None
@@ -502,12 +501,15 @@ class VideoWriter:
         return self.__writer
 
     @classmethod
-    def write_frames(cls, p, frames, fps=30, width=None, height=None):
-        with cls(p, width=width, height=height, fps=fps) as video:
-            for frame in frames:
-                if isinstance(frame, (str, cj.Path)):
-                    frame = cv2.imread(cj.Path(frame).path)
-                video.add_frame(frame)
+    def write_frames(cls, p, frames, fps=30, width=None, height=None, fourcc=None):
+        with cj.TempDir() as tmp_dir:
+            tmp_path = tmp_dir.path.join(cj.Path(p).name)
+            with cls(tmp_path.path, fourcc=fourcc, width=width, height=height, fps=fps) as video:
+                for frame in frames:
+                    if isinstance(frame, (str, cj.Path)):
+                        frame = cv2.imread(cj.Path(frame).path)
+                    video.add_frame(frame)
+            tmp_path.mv(p)
 
     def __enter__(self, *args, **kwargs):
         self._with_context = True
@@ -555,6 +557,7 @@ class _IVideo:
         pass
 
     @property
+    @abstractmethod
     def is_stream(self) -> bool:
         pass
 
@@ -756,7 +759,7 @@ class Screen(_IVideo):
         from mss import mss
         with mss() as sct:
             while self._capture:
-                yield Image(np.array(sct.grab(self._mon)), 'RGB').rgb_to_bgr().bgr_to_rgb().data
+                yield Image(np.array(sct.grab(self._mon)), 'RGBA').rgb_to_bgr()
 
     @property
     def total_frames(self) -> int:
@@ -784,18 +787,20 @@ class Screen(_IVideo):
 
 class Video:
 
-    def __init__(self, *args, fps=None, frame_preprocess=lambda frame: frame, **kwargs):
+    def __init__(self, *args, fps=None, frame_preprocess=None, **kwargs):
         kwargs['fps'] = fps
         self._speed = 1
-        self._frame_func_preprocess = frame_preprocess
+        self._frame_func_preprocess = frame_preprocess or (lambda x: x)
         self._args = args
         self._kwargs = kwargs
-        self._current_number_frame = 0
         self._build()
+        self._current_number_frame = 0
         self._th_show = None
         self._t0 = None
         self._fps_time = None
         self._count_frames = 0
+        self._th_show_running = False
+        self._last_frame = None
 
     def _build(self):
         if len(self._args):
@@ -826,8 +831,10 @@ class Video:
         assert hasattr(self, '_cap'), NotImplementedError(
                 f'Internal error. Please open new issue on https://github.com/cereja-project/calango')
         self._current_number_frame = 0
+        self._th_show = None
+        self._t0 = None
+        self._fps_time = None
         self._count_frames = 0
-
         self._th_show_running = False
         self._last_frame = None
 
@@ -858,8 +865,7 @@ class Video:
         self._current_number_frame += 1
         self._count_frames += 1  # for calculate fps correctly
         if self.current_number_frame > self.total_frames:
-            if isinstance(self._cap, cv2.VideoCapture):
-                self._cap.release()
+            self.stop()
             return None
         self._last_frame = image
         return image
@@ -947,13 +953,16 @@ class Video:
             raise NotImplementedError("Not implemented show video on colab")
         self._th_show_running = True
         try:
-            for image in self.get_frames():
+            while self.is_opened:
+                image = self.__get_next_frame()
+                if image is None:
+                    continue
                 cv2.imshow(self._cap.name, image.draw_text(self.video_info))
                 if self.is_break_view:
-                    self._cap.stop()
+                    self.stop()
         except Exception as e:
-            logging.error(e)
-            self._cap.stop()
+            logging.warning(e)
+            self.stop()
         self._th_show_running = False
 
     def get_frames(self, n_frames=None):
@@ -969,11 +978,14 @@ class Video:
             if n_frames is not None:
                 n_frames -= 1
             if n_frames == 0:
-                self._cap.stop()
+                self.stop()
                 break
 
-    def save(self, file_path, n_frames=None):
-        VideoWriter.write_frames(file_path, self.get_frames(n_frames=n_frames), fps=self._cap.fps)
+    def _save(self, file_path, n_frames=None, fourcc=None):
+        VideoWriter.write_frames(file_path, self.get_frames(n_frames=n_frames), fps=self._cap.fps, fourcc=fourcc)
+
+    def save(self, file_path, n_frames=None, fourcc=None):
+        threading.Thread(target=self._save, args=(file_path, n_frames, fourcc)).start()
 
     def show(self):
         if ON_COLAB_JUPYTER:
@@ -994,6 +1006,7 @@ class Video:
                 return
             if self._th_show is not None:
                 self._th_show.join()
+                self._build()
             self._th_show = threading.Thread(target=self._show, daemon=True)
             self._th_show.start()
 
@@ -1018,3 +1031,136 @@ class Video:
         self._fps_time = time.time()
         self._speed = speed
         self._count_frames = 0
+
+    def stop(self):
+        self._cap.stop()
+
+
+class VideoMagnification:
+    def __init__(self, *args, batch_size=20, levels=3, low=2.33, high=2.67, amplification=30, frame_preprocess=None,
+                 **kwargs):
+        self.video = Video(*args, frame_preprocess=frame_preprocess, **kwargs)
+        self.batch_size = batch_size
+        self.levels = levels
+        self.low = low
+        self.amplification = amplification
+        self.high = high
+
+    def get_batch_frames(self, kernel_size, strides=1):
+        batch_frames = []
+        batch_frames_tensor = []
+        while self.video.is_opened:
+            frame = self.video.next_frame
+            if frame is None:
+                continue
+            batch_frames_tensor.append(frame.copy())
+            laplacian = frame.laplacian_pyramid(self.levels)
+            if len(batch_frames) == 0:
+                for i in range(len(laplacian)):
+                    batch_frames.append([laplacian[i]])
+            else:
+                for i in range(len(laplacian)):
+                    batch_frames[i].append(laplacian[i])
+            if len(batch_frames_tensor) == kernel_size:
+                yield batch_frames_tensor.copy(), batch_frames.copy()
+                batch_frames = [pyramid_level[strides:] for pyramid_level in batch_frames]
+                batch_frames_tensor = batch_frames_tensor[strides:]
+
+    def butter_bandpass_filter(self, data, lowcut, highcut, fs, order=5):
+        from scipy import signal, fftpack
+        omega = 0.5 * fs
+        low = lowcut / omega
+        high = highcut / omega
+        b, a = signal.butter(order, [low, high], btype='band')
+        y = signal.lfilter(b, a, data, axis=0)
+        return y
+
+    def temporal_ideal_filter(self, tensor, low, high, fps, axis=0):
+        from scipy import signal, fftpack
+        fft = fftpack.fft(tensor, axis=axis)
+        frequencies = fftpack.fftfreq(len(tensor), d=1.0 / fps)
+        bound_low = (np.abs(frequencies - low)).argmin()
+        bound_high = (np.abs(frequencies - high)).argmin()
+        fft[:bound_low] = 0
+        fft[bound_high:-bound_high] = 0
+        fft[-bound_low:] = 0
+        iff = fftpack.ifft(fft, axis=axis)
+        return np.abs(iff)
+
+    def reconstract_from_tensorlist(self, filter_tensor_list):
+        final = np.zeros(filter_tensor_list[-1].shape)
+        for i in range(filter_tensor_list[0].shape[0]):
+            up = filter_tensor_list[0][i]
+            for n in range(len(filter_tensor_list) - 1):
+                k = filter_tensor_list[n + 1][i]
+                up = cv2.pyrUp(up, dstsize=(k.shape[1], k.shape[0])) + k
+            final[i] = up
+        return final
+
+    def magnify_motion(self, video_tensor, laplacian_tensor_list):
+        filter_tensor_list = []
+        for lap in laplacian_tensor_list:
+            filter_tensor = self.butter_bandpass_filter(lap, 0.4, 0.8, int(self.video.fps),
+                                                        order=len(laplacian_tensor_list))
+            filter_tensor *= 150
+            filter_tensor_list.append(filter_tensor)
+        recon = self.reconstract_from_tensorlist(filter_tensor_list)
+        final = video_tensor + recon
+
+        return final
+
+    def magnify_color(self, motion_video_tensor, laplacian_tensor_list):
+        output = self.gaussian_video(motion_video_tensor.copy(), len(laplacian_tensor_list))
+        output = self.temporal_ideal_filter(output, self.low, self.high, self.video.fps)[-1] * self.amplification
+        for n in range(len(laplacian_tensor_list) - 1):
+            output = cv2.pyrUp(output, dstsize=(
+                laplacian_tensor_list[n + 1][0].shape[1], laplacian_tensor_list[n + 1][0].shape[0]))
+        return motion_video_tensor[-1] + output
+
+    def build_gaussian_pyramid_last(self, src, level=3):
+        s = src.copy()
+        while level > 1:
+            s = cv2.pyrDown(s)
+            level -= 1
+        return s
+
+    def gaussian_video(self, video_tensor, levels=3):
+        vid_data = []
+        for frame in video_tensor.copy():
+            vid_data.append(self.build_gaussian_pyramid_last(frame, level=levels))
+        return np.array(vid_data)
+
+    @property
+    def frames(self):
+
+        for batch_frames_tensor, laplacian in self.get_batch_frames(self.batch_size):
+            mag_motion = self.magnify_motion(batch_frames_tensor, laplacian)
+            mag_color = self.magnify_color(motion_video_tensor=mag_motion, laplacian_tensor_list=laplacian)
+            mag_color = cv2.convertScaleAbs(mag_color)
+            yield cv2.convertScaleAbs(mag_motion[-1]), cv2.convertScaleAbs(batch_frames_tensor[-1]), mag_color
+
+    def show(self):
+        with VideoWriter('./magnify.mp4', fps=self.video.fps, ) as writer:
+
+            try:
+                for n, (frame, original, color) in enumerate(self.frames):
+                    h, w, z = frame.shape
+                    mask = Image.get_empty_image()
+                    mask.top.left = Image(frame).write_text(f'magnified {(self.low, self.high, self.amplification)}')
+                    mask.bottom = Image(original).write_text('Original')
+                    mask.top.right = Image(color).write_text('Color')
+                    writer.add_frame(mask)
+                    cv2.imshow("Video", Image(mask).write_text(f'Frame: {n + 1} - FPS {self.video.fps}'))
+                    if self.video.is_break_view:
+                        break
+            finally:
+                self.video.stop()
+                cv2.destroyAllWindows()
+        # try:
+        #     for n, frame in enumerate(self.frames):
+        #         cv2.imshow("Video", Image(frame).write_text(f'Frame: {n + 1} - FPS {self.video.fps}'))
+        #         if self.video.is_break_view:
+        #             break
+        # finally:
+        #     self.video.stop()
+        #     cv2.destroyAllWindows()
